/*
 * Copyright (C) 2011 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michal Hruby <michal.hruby@canonical.com>
 *
 */
/**
 * SECTION:dee-client
 * @short_description: Creates a client object you can use to connect
 *                     to DeeServer
 * @include: dee.h
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gio/gio.h>

#include "dee-client.h"
#include "dee-marshal.h"
#include "trace-log.h"

G_DEFINE_TYPE (DeeClient, dee_client, DEE_TYPE_PEER)

#define GET_PRIVATE(o) \
      (G_TYPE_INSTANCE_GET_PRIVATE ((o), DEE_TYPE_CLIENT, DeeClientPrivate))

/**
 * DeeClientPrivate:
 *
 * Ignore this structure.
 **/
struct _DeeClientPrivate
{
  GDBusConnection *connection;
  GCancellable    *cancellable;
};

/* Globals */
enum
{
  PROP_0
};

enum
{
  LAST_SIGNAL
};

//static guint32 _server_signals[LAST_SIGNAL] = { 0 };

/* Forwards */
static gboolean     dee_client_is_swarm_leader     (DeePeer *peer);
static const gchar* dee_client_get_swarm_leader    (DeePeer *peer);
static void         connection_finished            (GObject *object,
                                                    GAsyncResult *res,
                                                    gpointer user_data);

/* GObject methods */
static void
dee_client_get_property (GObject *object, guint property_id,
                         GValue *value, GParamSpec *pspec)
{
  switch (property_id)
    {
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
}

static void
dee_client_set_property (GObject *object, guint property_id,
                              const GValue *value, GParamSpec *pspec)
{
  switch (property_id)
    {
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
}

static void
dee_client_constructed (GObject *self)
{
  DeeClientPrivate *priv;
  gchar            *bus_address;
  const gchar      *swarm_name;
  GDBusConnectionFlags flags;

  priv = DEE_CLIENT (self)->priv;

  /* we should chain up the constructed method here, but peer does things we
   * don't want to, so not chaining up... */

  swarm_name = dee_peer_get_swarm_name (DEE_PEER (self));
  if (swarm_name == NULL)
    {
      g_critical ("DeeClient created without a swarm name. You must specify "
                  "a non-NULL swarm name");
      return;
    }

  bus_address = g_strdup_printf ("unix:abstract=%s", swarm_name);

  flags = G_DBUS_CONNECTION_FLAGS_AUTHENTICATION_CLIENT;
  priv->cancellable = g_cancellable_new ();
  g_dbus_connection_new_for_address (bus_address,
                                     flags,
                                     NULL, // AuthObserver
                                     priv->cancellable,
                                     connection_finished,
                                     self);

  g_free (bus_address);
}

static void
dee_client_finalize (GObject *object)
{
  DeeClientPrivate *priv;

  priv = DEE_CLIENT (object)->priv;

  if (priv->cancellable)
    {
      g_cancellable_cancel (priv->cancellable);
      g_object_unref (priv->cancellable);
    }

  if (priv->connection)
    {
      // FIXME: close the connection
      g_object_unref (priv->connection);
    }

  G_OBJECT_CLASS (dee_client_parent_class)->finalize (object);
}

static void
dee_client_class_init (DeeClientClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  DeePeerClass *peer_class = DEE_PEER_CLASS (klass);

  g_type_class_add_private (klass, sizeof (DeeClientPrivate));

  object_class->constructed    = dee_client_constructed;
  object_class->get_property   = dee_client_get_property;
  object_class->set_property   = dee_client_set_property;
  object_class->finalize       = dee_client_finalize;

  peer_class->is_swarm_leader  = dee_client_is_swarm_leader;
  peer_class->get_swarm_leader = dee_client_get_swarm_leader;
}

static void
dee_client_init (DeeClient *self)
{
  self->priv = GET_PRIVATE (self);
}

DeeClient*
dee_client_new (const gchar* swarm_name)
{
  g_return_val_if_fail (swarm_name != NULL, NULL);

  return g_object_new (DEE_TYPE_CLIENT, "swarm-name", swarm_name, NULL);
}

/* Private Methods */

static gboolean
dee_client_is_swarm_leader (DeePeer *peer)
{
  return FALSE;
}

static const gchar*
dee_client_get_swarm_leader (DeePeer *peer)
{
  DeeClientPrivate *priv;

  priv = DEE_CLIENT (peer)->priv;
  return priv->connection ? g_dbus_connection_get_guid (priv->connection) : NULL;
}

static gboolean
emit_peer_found (gpointer user_data)
{
  g_return_val_if_fail (DEE_IS_CLIENT (user_data), FALSE);

  DeeClientPrivate *priv = DEE_CLIENT (user_data)->priv;

  g_signal_emit_by_name (user_data, "peer-found",
                         g_dbus_connection_get_guid (priv->connection));

  return FALSE;
}

static void
connection_finished (GObject *object, GAsyncResult *res, gpointer user_data)
{
  GDBusConnection  *connection;
  DeeClientPrivate *priv;
  GError           *error = NULL;

  connection = g_dbus_connection_new_for_address_finish (res, &error);

  if (error)
    {
      if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
        {
          g_warning ("Unable to connect to server: %s", error->message);
        }
      else
        {
          // swarm-leader will be set to NULL for unsuccessful connections
          g_object_notify (G_OBJECT (user_data), "swarm-leader");
        }

      g_error_free (error);
      return;
    }

  priv = DEE_CLIENT (user_data)->priv;
  priv->connection = connection;

  g_object_unref (priv->cancellable);
  priv->cancellable = NULL;

  g_object_notify (G_OBJECT (user_data), "swarm-leader");

  // FIXME: we might want to call some List method (same as DeePeer), so far
  // we'll just simulate an async method
  g_idle_add (emit_peer_found, user_data);
}

