/*
 * Copyright (C) 2011 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by: Michal Hruby <michal.hruby@canonical.com>
 *
 */
/**
 * SECTION:dee-server
 * @short_description: Creates a server object you can connect to.
 * @include: dee.h
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gio/gio.h>

#include "dee-server.h"
#include "dee-marshal.h"
#include "trace-log.h"

G_DEFINE_TYPE (DeeServer, dee_server, DEE_TYPE_PEER)

#define GET_PRIVATE(o) \
      (G_TYPE_INSTANCE_GET_PRIVATE ((o), DEE_TYPE_SERVER, DeeServerPrivate))

/**
 * DeeServerPrivate:
 *
 * Ignore this structure.
 **/
struct _DeeServerPrivate
{
  GDBusServer *server;

  GSList *active_connections;
};

/* Globals */
enum
{
  PROP_0
};

enum
{
  LAST_SIGNAL
};

//static guint32 _server_signals[LAST_SIGNAL] = { 0 };

/* Forwards */
static gboolean on_new_connection              (GDBusServer *server,
                                                GDBusConnection *connection,
                                                gpointer user_data);

static void     on_connection_closed           (GDBusConnection *connection,
                                                gboolean remote_peer_vanished,
                                                GError *error,
                                                gpointer user_data);

static gboolean dee_server_is_swarm_leader     (DeePeer *peer);

/* GObject methods */
static void
dee_server_get_property (GObject *object, guint property_id,
                         GValue *value, GParamSpec *pspec)
{
  switch (property_id)
    {
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
}

static void
dee_server_set_property (GObject *object, guint property_id,
                              const GValue *value, GParamSpec *pspec)
{
  switch (property_id)
    {
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
}

static void
dee_server_constructed (GObject *self)
{
  DeeServerPrivate *priv;
  GError           *error = NULL;
  gchar            *bus_address;
  gchar            *guid;
  const gchar      *swarm_name;
  GDBusServerFlags  server_flags;

  priv = DEE_SERVER (self)->priv;

  /* we should chain up the constructed method here, but peer does things we
   * don't want to, so no chaining up... */

  swarm_name = dee_peer_get_swarm_name (DEE_PEER (self));
  if (swarm_name == NULL)
    {
      g_critical ("DeeServer created without a swarm name. You must specify "
                  "a non-NULL swarm name");
      return;
    }

  bus_address = g_strdup_printf ("unix:abstract=%s", swarm_name);
  guid = g_dbus_generate_guid ();
  server_flags = G_DBUS_SERVER_FLAGS_AUTHENTICATION_ALLOW_ANONYMOUS;

  priv->server = g_dbus_server_new_sync (bus_address,
                                         server_flags,
                                         guid,
                                         NULL,
                                         NULL,
                                         &error);

  if (error)
    {
      g_critical ("Unable to set up DBusServer: %s", error->message);

      g_error_free (error);
      g_free (bus_address);
      g_free (guid);
      return;
    }

  g_signal_connect (priv->server, "new-connection",
                    G_CALLBACK (on_new_connection), self);

  g_dbus_server_start (priv->server);

  g_free (bus_address);
  g_free (guid);
}

static void
close_connection (gpointer data, gpointer user_data)
{
  GDBusConnection *connection = G_DBUS_CONNECTION (data);

  g_signal_handlers_disconnect_by_func (connection, on_connection_closed,
                                        user_data);

  // FIXME: should we use the sync variant? and flush first?
  g_dbus_connection_close (connection, NULL, NULL, NULL);
}

static void
dee_server_finalize (GObject *object)
{
  DeeServerPrivate *priv;

  priv = DEE_SERVER (object)->priv;

  if (priv->active_connections)
    {
      g_slist_foreach (priv->active_connections, close_connection, object);
      g_slist_free_full (priv->active_connections, g_object_unref);
      priv->active_connections = NULL;
    }

  if (priv->server)
    {
      /* this should be done automatically on unref, but it doesn't seem so */
      g_dbus_server_stop (priv->server);

      g_object_unref (priv->server);
    }

  G_OBJECT_CLASS (dee_server_parent_class)->finalize (object);
}

static void
dee_server_class_init (DeeServerClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  DeePeerClass *peer_class = DEE_PEER_CLASS (klass);

  g_type_class_add_private (klass, sizeof (DeeServerPrivate));

  object_class->constructed = dee_server_constructed;
  object_class->get_property = dee_server_get_property;
  object_class->set_property = dee_server_set_property;
  object_class->finalize = dee_server_finalize;

  peer_class->is_swarm_leader = dee_server_is_swarm_leader;
}

static void
dee_server_init (DeeServer *self)
{
  self->priv = GET_PRIVATE (self);
}

DeeServer*
dee_server_new (const gchar* swarm_name)
{
  g_return_val_if_fail (swarm_name != NULL, NULL);

  return g_object_new (DEE_TYPE_SERVER, "swarm-name", swarm_name, NULL);
}

const gchar*
dee_server_get_client_address (DeeServer *server)
{
  DeeServerPrivate *priv;
  g_return_val_if_fail (DEE_IS_SERVER (server), NULL);

  priv = server->priv;

  return g_dbus_server_get_client_address (priv->server);
}

/* Private Methods */
static gboolean
on_new_connection (GDBusServer *server,
                   GDBusConnection *connection,
                   gpointer user_data)
{
  DeeServer *self = DEE_SERVER (user_data);
  DeeServerPrivate *priv = self->priv;

  trace_object (self, "New connection: [%p]", connection);

  // FIXME: check same user_id?

  priv->active_connections = g_slist_prepend (priv->active_connections,
                                              g_object_ref (connection));

  g_signal_connect (connection, "closed",
                    G_CALLBACK (on_connection_closed), self);

  // FIXME: time to register dbus objects on this connection

  return TRUE;
}

static void
on_connection_closed (GDBusConnection *connection,
                      gboolean remote_peer_vanished,
                      GError *error,
                      gpointer user_data)
{
  GSList           *element;
  DeeServer        *self = DEE_SERVER (user_data);
  DeeServerPrivate *priv = self->priv;

  trace_object (self, "Connection [%p] was closed", connection);

  element = g_slist_find (priv->active_connections, connection);
  if (element == NULL)
    {
      g_warning ("Connection closed for element which isn't "
                 "in active_connections");
      return;
    }

  g_object_unref (connection);
  priv->active_connections = g_slist_delete_link (priv->active_connections,
                                                  element);
}

static gboolean
dee_server_is_swarm_leader (DeePeer *peer)
{
  DeeServerPrivate *priv = DEE_SERVER (peer)->priv;
  return priv->server != NULL;
}

