/*
 * Copyright (C) 2011 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as 
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by
 *              Mikkel Kamstrup Erlandsen <mikkel.kamstrup@canonical.com>
 *
 */

#include "config.h"
#include <math.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <glib-object.h>

#include <dee.h>

typedef struct _Benchmark Benchmark;
typedef void (*BenchmarkFunc) (Benchmark *benchmark);
typedef void (*BenchmarkSetup) (Benchmark *benchmark);
typedef void (*BenchmarkTeardown) (Benchmark *benchmark);

typedef struct {
  gdouble elapsed;
} RunData;

struct _Benchmark {
  const gchar       *name;
  BenchmarkSetup     benchmark_setup;
  BenchmarkFunc      benchmark_func;
  BenchmarkTeardown  benchmark_teardown;
  guint              n_runs;
  RunData           *runs;
  gpointer           state;
};

static GList *benchmarks = NULL;

static void
add_benchmark (Benchmark *benchmark)
{
  benchmarks = g_list_append (benchmarks, benchmark);
}

static void
run_benchmark (Benchmark *bench)
{
  GTimer *timer;
  guint   i;
  gdouble total_runtime, avg_runtime, std_dev;
  
  bench->runs = g_new0 (RunData, bench->n_runs + 1);
  timer = g_timer_new ();
  total_runtime = 0;
  
  bench->benchmark_setup (bench);
  
  g_printf ("=== %s ===\n", bench->name);
      
  for (i = 0; i < bench->n_runs; i++)
  {
    g_timer_start (timer);
    bench->benchmark_func (bench);
    bench->runs[i].elapsed = g_timer_elapsed (timer, NULL);
    total_runtime += bench->runs[i].elapsed;
  }
  
  /* Compute average runtime */
  avg_runtime = total_runtime / bench->n_runs;
  
  /* Compute standard deviation */
  std_dev = 0;
  for (i = 0; i < bench->n_runs; i++)
    {
      gdouble delta = bench->runs[i].elapsed - avg_runtime;
      std_dev += delta * delta;
    }
  std_dev = sqrt (std_dev / bench->n_runs);
  
  /* Print report */
  g_printf ("Runs           : %u\n", bench->n_runs);
  g_printf ("Total runtime  : %fs\n", total_runtime);
  g_printf ("Avg. runtime   : %fs\n", avg_runtime);
  g_printf ("Std. deviation : %fs\n", std_dev);
  g_printf ("\n");
  
  
  
  g_timer_destroy (timer);
  bench->benchmark_teardown (bench);
  // purposely leak bench->runs. Caller may want it
}

static void
run_benchmarks (void)
{
  GList  *iter;
  
  for (iter = benchmarks; iter; iter = iter->next)
    {
      Benchmark *bench = (Benchmark*) iter->data;
      run_benchmark (bench);
    }
}

static void
bench_seqmodel_setup (Benchmark *bench)
{
  DeeModel *model;
  
  model = dee_sequence_model_new ();
  dee_model_set_schema (model, "s", "s", "s", "u", "b", NULL);
  bench->state = model;
}

static void
bench_seqmodel_append_run (Benchmark *bench)
{
  DeeModel *model;
  guint     limit = 1000, i;
  
  g_assert (DEE_IS_MODEL (bench->state));
  
  model = DEE_MODEL (bench->state);
  
  for (i = 0; i < limit; i++)
    {
      dee_model_append (model, "Hello", "world", "!", 42, TRUE);
    }
}

static void
bench_seqmodel_prepend_run (Benchmark *bench)
{
  DeeModel *model;
  guint     limit = 1000, i;
  
  g_assert (DEE_IS_MODEL (bench->state));
  
  model = DEE_MODEL (bench->state);
  
  for (i = 0; i < limit; i++)
    {
      dee_model_prepend (model, "Hello", "world", "!", 42, TRUE);
    }
}

static void
bench_gobject_teardown (Benchmark *bench)
{
  GObject *obj;
  
  obj = G_OBJECT (bench->state);
  g_object_unref (obj);
}

gint
main (gint argc, gchar *argv[])
{
  g_type_init ();
  g_thread_init (NULL);

  Benchmark seqmodel_append = { "SequenceModel.append()",
                                bench_seqmodel_setup,
                                bench_seqmodel_append_run,
                                bench_gobject_teardown,
                                10,
                                NULL };
  add_benchmark (&seqmodel_append);
  
  Benchmark seqmodel_prepend = { "SequenceModel.prepend()",
                                bench_seqmodel_setup,
                                bench_seqmodel_prepend_run,
                                bench_gobject_teardown,
                                10,
                                NULL };
  add_benchmark (&seqmodel_prepend);

  run_benchmarks ();
  return 0;
}
